CastTimeAnalysis = LibStub("AceAddon-3.0"):NewAddon("CastTimeAnalysis", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0");
local core = CastTimeAnalysis;
local L = LibStub("AceLocale-3.0"):GetLocale("CastTimeAnalysis")

--[[local options = {
	name = "CastTimeAnalysis",
	type = 'group',
	args = {
	},
};]]
local defaults = {
	profile = {
		bossOnly = true,
		fullCombat = true,
	},
};

--[[
	known bugs:
	- stopping a cast but having it succeed is marked as a wasted + gcd
		UNIT_SPELLCAST_STOP player Fireball (1)
		UNIT_SPELLCAST_SENT player Fireball (2)
		UNIT_SPELLCAST_SUCCEEDED player Fireball (1)
		UNIT_SPELLCAST_START player Fireball (2)
]]
local OOC_THRESH = 2; -- time you need to be out of combat to actually be considered done
local USED_TYPE = {
	["channel"]	= true,
	["global"]	= true,
	["cast"]	= true,
};
local QUIET_TYPE = {
	["gcd"] = true,
	["wasted"] = true,
}
local debugFlag = false;

core.debug = 0;
--[===[@debug@
core.debug = 1; -- debug printing by level
--@end-debug@]===]
core.combat = false; -- current state of combat
core.lastCombat = 0; -- time you last left combat
core.oocTimer = false; -- timer used to print stats after OOC_THRESH seconds
core.currentBoss = false;
core.current = {
	type = "quiet",	-- current spell's type (quiet | lag | gcd | channel | cast)
	name = "",
	start = 0, -- current spell's start time
	last = 0, -- last recorded segment (for undoing wasted on cancelled successful casts)
	gcdUntil = 0, -- timestamp when gcd will be off for fixing extra lag occurences
};
core.stats = {
	start	= 0, -- time you started combat
	length	= 0, -- time spent in combat
	total	= 0, -- length of the fight (via adding fields below)
	lag		= 0, -- time lost between CastSent and CastStart
	wasted	= 0, -- cast time lost to interrupts
	quiet	= 0, -- time spent not doing anything
	dead	= 0, -- time spent dead
	used	= 0, -- total time spent casting
	channel	= 0, -- time spent channeling
	global	= 0, -- time spent in globals
	cast	= 0, -- time spent in casts
};

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("CastTimeAnalysisDB", defaults, "Default");
	self:RegisterChatCommand("cta", "SlashHandle");
	self:RegisterChatCommand("casttimeanalysis", "SlashHandle");
	--[[
		LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("", options);
		local ACD = LibStub("AceConfigDialog-3.0");
		ACD:AddToBlizOptions("", "");
		self:SetEnabledState(self.db.profile.enabled;
	]]
end

function core:OnEnable()
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "EnterCombat");
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "LeaveCombat");
	
	self:RegisterEvent("UNIT_SPELLCAST_SENT", "CastSent");
	
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "ChannelStart");
	self:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "ChannelStop");
	
	self:RegisterEvent("UNIT_SPELLCAST_START", "CastStart");
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "CastFinish");
	
	self:RegisterEvent("UNIT_SPELLCAST_STOP", "CastStop");
	self:RegisterEvent("UNIT_SPELLCAST_FAILED", "CastStop");
	
	self:RegisterEvent("PLAYER_DEAD", "DeathStart");
	self:RegisterEvent("PLAYER_ALIVE", "DeathStop");
	
	if (self.db.profile.bossOnly) then
		self:RegisterEvent("UNIT_TARGET", "FindBoss");
	end
	--self:RegisterEvent("UNIT_SPELLCAST_FAILED_QUIET", "CastStop"); -- this seems to spam every time you unsuccessfully cast
end

function core:OnDisable()
	
end

function core:SlashHandle(message)
	message = string.lower(message);
	if (message == "bossonly" or message == "bo") then
		self.db.profile.bossOnly = not self.db.profile.bossOnly;
		if (self.db.profile.bossOnly) then
			self:Print(L["Reporting enabled for boss fights only."]);
			self:UnregisterEvent("UNIT_TARGET");
		else 
			self:Print(L["Reporting enabled for all fights."]);
			self:RegisterEvent("UNIT_TARGET", "FindBoss");			
		end
	elseif (message == "raidcombat" or message == "rc") then
		self.db.profile.fullCombat = not self.db.profile.fullCombat;
		if (self.db.profile.fullCombat) then
			self:Print(L["Reporting will occur after the entire raid leaves combat."]);
		else 
			self:Print(L["Reporting will occur after only the player leaves combat."]);
		end
	else
		self:BlankPrint(format(L["%s Options"], "|cfff88017Cast Time Analysis|r"));
		self:BlankPrint("  |cfff88017bossonly|r|||cfff88017bo|r - " .. self:GetSwitch("bossOnly") .. " - " .. L["Toggles reporting only after boss fights"]);
		self:BlankPrint("  |cfff88017raidcombat|r|||cfff88017rc|r - " .. self:GetSwitch("fullCombat") .. " - " .. L["Toggles reporting after the player or entire raid leaves combat"]);
	end
end

function core:GetSwitch(option)
	if (self.db.profile[option]) then
		return "[|cff00ff00" .. L["ON"] .. "|r]";
	end
	
	return "[|cffff0000" .. L["OFF"] .. "|r]";
end

-- flow:
--  Expect() - expect a certain current state
--  Report() - adds to stats based on running current
--  Update() - changes current to new value

function core:Expect(type, spell, event)
	-- given: expected type[, spell]
	-- compare to self.current values
	-- wasted is synonymous to quiet for all cases where quiet is expected
	local curType = (self.current.type and QUIET_TYPE[self.current.type] and "quiet") or self.current.type;
	if (curType ~= type or (spell and self.current.spell ~= spell)) then
		--[[if (not silent) then
			self:AssertFailed(event, self.current, {type=type, spell=spell});
		end]]
		return false;
	end
	return true;
end

function core:Report(type, time)
	-- debug:
	if (self.current.type == "gcd") then
		self:Debug(2, "GCD Quiet:", GetTime()-self.current.start);
	end
	
	-- [type] is retrieved from current state if not provided
	local curType = type or (self.current.type and QUIET_TYPE[self.current.type] and "quiet") or self.current.type;
	
	-- in-gcd lag procedures only apply for non-quiet states
	if (curType ~= "quiet") then
		-- if we're reporting before gcd is up, report as quiet
		if (GetTime() < self.current.gcdUntil) then
			self:Report("quiet");
			return;
		end
		
		-- if event started before gcd was up, make sure that extra goes to quiet
		if (not time and self.current.start < self.current.gcdUntil) then
			self:Report("quiet", self.current.gcdUntil-self.current.start);
			self.current.start = self.current.gcdUntil;
		end
	end
	
	local t = time or GetTime() - self.current.start;
	
	self.stats.total = self.stats.total + t;
	self.stats[curType] = self.stats[curType] + t;
	if (USED_TYPE[curType]) then 
		self.stats.used = self.stats.used + t;
	end	
	self.current.last = t;
end

function core:Update(type, spell)
	if (debugFlag) then
		debugFlag = false;
		self:Debug(2, type, spell);
	end
	if (type == "gcd") then
		-- special recording
		self.current.gcdUntil = GetTime()+spell;
		spell = nil;
	end
	-- given: type[, spell]
	self.current.type = type;
	self.current.spell = spell;
	self.current.start = GetTime();
end

function core:EnterCombat()
	if (GetTime()-self.lastCombat < OOC_THRESH) then
		-- got pulled back in
		self:CancelTimer(self.oocTimer, true);
		self.oocTimer = false;
		self.combat = true;
		return;
	end
	
	-- wipe current stats
	for k,v in pairs(self.stats) do
		self.stats[k] = 0;
	end
	
	self.combat = true;
	self.stats.start = GetTime();
	self.currentBoss = false;
	if (self.db.profile.bossOnly) then
		self:RegisterEvent("UNIT_TARGET", "FindBoss");
	end
	if (UnitCastingInfo("player")) then
		self:Update("cast", UnitCastingInfo("player"));
	elseif (UnitChannelInfo("player")) then
		self:Update("channel", UnitChannelInfo("player"));
	else
		self:Update("quiet");
	end
end

function core:LeaveCombat()
	self:Debug(2, "Combat Leave Check");
	if (UnitAffectingCombat("player")) then
		return;
	end
	
	-- check built for fullCombat option
	if (self.db.profile.fullCombat) then
		for i=1,GetNumRaidMembers() do
			if (UnitAffectingCombat("raid"..i)) then
				self:ScheduleTimer("LeaveCombat", 0.1);
				return;
			end
		end
	end
	
	self:Debug(1, "Combat Leave Success");
	self:FindBoss();
	
	if (self:Expect("channel") or self:Expect("dead")) then
		self:Report();
	elseif (self:Expect("cast", nil, "LEAVE_COMBAT")) then
		self:Report("wasted");
	else
		self:Report("quiet");
	end
	self.combat = false;
	self.lastCombat = GetTime();
	self.stats.length = GetTime()-self.stats.start;
	self.oocTimer = self:ScheduleTimer("ReportStats", OOC_THRESH);
end

function core:CastSent(e, uid, spell)
	if (uid ~= "player" or not self.combat) then return; end
	self:Debug(3, e, uid, spell);
	if (self:Expect("quiet")) then
		self:Report("quiet"); -- must specify "quiet" here because it could have compared to "wasted"
		self:Update("lag", spell);
	elseif (self:Expect("cast") or self:Expect("channel")) then
		-- in the middle of another cast
	else
		-- self:AssertFailed(e, "quiet", "cast", "channel");
		-- spells can be sent at any time, so no reason to assert
	end
end

function core:ChannelStart(e, uid, spell)
	if (uid ~= "player" or not self.combat) then return; end
	self:Debug(3, e, uid, spell);
	if (self:Expect("lag") or self:Expect("quiet")) then
		self:Report("lag");
		self:Update("channel", spell);
	else
		self:AssertFailed(e, {"lag", spell});
	end
end

function core:ChannelStop(e, uid, spell)
	if (uid ~= "player" or not self.combat) then return; end
	self:Debug(3, e, uid, spell);
	if (self:Expect("channel", spell)) then
		self:Report();
		self:Update("quiet");
	else
		self:AssertFailed(e, {"channel", spell});
	end
end

function core:CastStart(e, uid, spell)
	if (uid ~= "player" or not self.combat) then return; end
	self:Debug(3, e, uid, spell);
	if (self:Expect("lag") or self:Expect("quiet")) then
		-- if we're in quiet, it's because a STOP was fired after a SENT, a side-effect of the cast change a while ago
		self:Report("lag");
		self:Update("cast", spell);
	else
		self:AssertFailed(e, {"lag", spell}, "quiet");
	end
end

function core:CastFinish(e, uid, spell)
	if (uid ~= "player" or not self.combat) then return; end
	self:Debug(3, e, uid, spell);
	if (self:Expect("wasted", spell)) then
		-- special case; need to undo the wasted and mark it as a success
		local last = self.current.last;
		self:Debug(2, "wasted", last);
		self.stats.wasted = self.stats.wasted - last;
		self.stats.cast = self.stats.cast + last;
		self.stats.used = self.stats.used + last;
		self:Report("cast");
		self:Update("quiet");
	elseif (self:Expect("cast", spell)) then
		-- cast finished
		self:Report();
		self:Update("quiet");
	elseif (self:Expect("lag") or self:Expect("quiet")) then	
		-- if we're in quiet, it's because a STOP was fired after a SENT, a side-effect of the cast change a while ago
		-- gcd
		local gcd = select(2, GetSpellCooldown(spell)) or 0;
		if (gcd > 0.25) then
			-- when using instant spells that are not on GCD or on-use trinkets/enchants, gcd should report as empty or 0.0001
			-- hopefully no spell legitimately goes below 0.25s gcd
			self:Debug(2, "GCD:", gcd);
			self:Report("lag");
			-- lag reported, now report gcd; this will be subtracted from quiet later
			self:Report("global", select(2, GetSpellCooldown(spell)));
			self:Update("gcd", gcd);
		end
	elseif (self:Expect("channel", spell)) then
		-- extra event, we already know the channel started
	else
		self:AssertFailed(e, {"cast", spell}, {"lag", spell}, {"channel", spell}, "quiet");
	end
end

function core:CastStop(e, uid, spell)
	if (uid ~= "player" or not self.combat) then return; end
	self:Debug(3, e, uid, spell);
	if (self:Expect("cast", spell)) then
		self:Report("wasted");
		self:Update("quiet");
	else
		-- self:AssertFailed(e, {"cast", spell}, "quiet");
		-- we can get UNIT_SPELLCAST_FAILED at any time, so it is pointless to assert
	end
end

function core:DeathStart()
	if (self:Expect("cast")) then
		self:Report("wasted");
	else
		self:Report();
	end
	self:Update("dead");
end

function core:DeathStop()
	self:Report("dead");
	self:Update("quiet");
end

function core:FindBoss(_, unit)
	self:Debug(2, "FindBoss", unit, self.currentBoss)
	if (not self.currentBoss) then
		unit = (unit and unit.."target") or "target";
		-- check if it's a boss, and then check to make sure the boss itself is in combat (note that dummys never actually enter combat)
		if (UnitClassification(unit) == "worldboss" and (UnitAffectingCombat(unit) or UnitName(unit) == L["Heroic Training Dummy"])) then
			self.currentBoss = UnitName(unit);
			self:UnregisterEvent("UNIT_TARGET");
		end
	end
end

function core:ReportStats()
	-- adjust quiet for gcd
	self.stats.quiet = self.stats.quiet - self.stats.global; -- max(0, self.stats.quiet - self.stats.global);
	self.stats.total = self.stats.total - self.stats.global; -- max(0, self.stats.total - self.stats.global);
	self:Debug(1, "Reporting: ", self.db.profile.bossOnly, self.currentBoss);
	-- output
	if (not self.db.profile.bossOnly or self.currentBoss) then
		local combatTime = self.stats.length;
		local m, s = floor(combatTime/60), combatTime%60;
		local tab = "   ";
		local headFormat = (self.currentBoss and L["%s Combat Report for %s"]) or L["%s Combat Report"];
		self:BlankPrint(format(headFormat, "|cfff88017Cast Time Analysis|r", self.currentBoss));-- format("In combat for %d:%02d (%.1fs)", m, s, combatTime, self.stats.total));
		self:BlankPrint(self:GetStat("total", nil, 1));
		if (self.stats.dead > 0) then
			self:BlankPrint(self:GetStat("dead", "total", 2));
		end
		self:BlankPrint(self:GetStat("quiet", "total", 2));
		self:BlankPrint(self:GetStat("lag", "total", 2));
		self:BlankPrint(self:GetStat("wasted", "total", 2));
		self:BlankPrint(self:GetStat("used", "total", 2));
		self:BlankPrint(self:GetStat("channel", "used", 3));
		self:BlankPrint(self:GetStat("global", "used", 3));
		self:BlankPrint(self:GetStat("cast", "used", 3));
	end
end

function core:BlankPrint(...)
	print(...);
end

function core:GetStat(key, denom, tabs)
	return format("%s%s: %.1fs%s", ("      "):rep(tabs), L[key:sub(1, 1):upper() .. key:sub(2)], self.stats[key], (denom and format(" (%.1f%%)", self.stats[key]/max(0.1, self.stats[denom])*100)) or "");
end

function core:AssertFailed(event, ...)
	local given = self.current;
	local expected = "";
	for i=1, select("#", ...) do
		local cur = select(i, ...);
		if (type(cur) == "table") then
			cur = format("%s(%s)", cur[1], cur[2]);
		else
			cur = format("%s()", cur);
		end
		
		expected = format("%s|%s", cur, expected);
	end
	
	self:Debug(1, format("Assertion failed on %s. Given:%s(%s); Expected:%s", event, given.type or "", given.spell or "", expected:sub(1, -2)));
end

function core:ScheduleGCD()
	
end

function core:Debug(level, ...)
	if (type(level) ~= "number") then
		self:Debug(1, level, ...);
		return;
	end
	if self.debug >= level then 
		self:Print(...);
	end
end